  `/*******************************************************************************
* FILE NAME: motor_routines.c
*
* DESCRIPTION:
*  This file contains motor/joystick related filters.
*
* USAGE:
*
*******************************************************************************/

#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "timers.h"

#include "user_routines.h"
#include "clock_routines.h"
#include "motor_routines.h"
#include "print_routines.h"
#include "auto_routines.h"

#ifndef _FRC_BOARD
  unsigned char p1_y,p2_y,p3_y,p4_y,p1_x,p2_x,p3_x,p4_x,p1_wheel,p2_wheel,p3_wheel,p4_wheel;
  unsigned char pwm09,pwm10,pwm11,pwm12,pwm13,pwm14,pwm15,pwm16;
#endif

/*******************************************************************************
* FUNCTION NAME: motor_fix_in
* PURPOSE:       Converts pwm 0 - 254 range to -127 to 127
 *******************************************************************************/
void motor_fix_in(void)
{
	ip1_y = (signed int) p1_y - 127;
	ip2_y = (signed int) p2_y - 127;
	ip3_y = (signed int) p3_y - 127;
	ip4_y = (signed int) p4_y - 127;
	ip1_x = (signed int) p1_x - 127;
	ip2_x = (signed int) p2_x - 127;
	ip3_x = (signed int) p3_x - 127;
	ip4_x = (signed int) p4_x - 127;
    ip1_wheel = (signed int) p1_wheel - 127;
    ip2_wheel = (signed int) p2_wheel - 127;   
    ip3_wheel = (signed int) p3_wheel - 127;   
    ip4_wheel = (signed int) p4_wheel - 127; 
}

/*******************************************************************************
* FUNCTION NAME: motor_fix_out
* PURPOSE:       Converts pwm -127 to 127 range to 0 - 254
*******************************************************************************/
void motor_fix_out(void)
{
  if (ipwm01 == -128) ipwm01--;
  if (ipwm02 == -128) ipwm01--;
  if (ipwm03 == -128) ipwm01--;
  if (ipwm04 == -128) ipwm01--;
  if (ipwm05 == -128) ipwm01--;
  if (ipwm06 == -128) ipwm01--;
  if (ipwm07 == -128) ipwm01--;
  if (ipwm08 == -128) ipwm01--;
  if (ipwm09 == -128) ipwm01--;
  if (ipwm10 == -128) ipwm01--;
  if (ipwm11 == -128) ipwm01--;
  if (ipwm12 == -128) ipwm01--;
  if (ipwm13 == -128) ipwm01--;
  if (ipwm14 == -128) ipwm01--;
  if (ipwm15 == -128) ipwm01--;
  if (ipwm16 == -128) ipwm01--;

	pwm01 = (unsigned char) ipwm01 +127;
	pwm02 = (unsigned char) ipwm02 +127;
	pwm03 = (unsigned char) ipwm03 +127;
	pwm04 = (unsigned char) ipwm04 +127;
	pwm05 = (unsigned char) ipwm05 +127;
	pwm06 = (unsigned char) ipwm06 +127;
	pwm07 = (unsigned char) ipwm07 +127;
	pwm08 = (unsigned char) ipwm08 +127;
	pwm09 = (unsigned char) ipwm09 +127;
	pwm10 = (unsigned char) ipwm10 +127;
	pwm11 = (unsigned char) ipwm11 +127;
	pwm12 = (unsigned char) ipwm12 +127;
	pwm13 = (unsigned char) ipwm13 +127;
	pwm14 = (unsigned char) ipwm14 +127;
	pwm15 = (unsigned char) ipwm15 +127;
	pwm16 = (unsigned char) ipwm16 +127;
}



/*******************************************************************************
* FUNCTION NAME: Limit_Mix
* PURPOSE:       Limits the mixed value for one joystick drive.
* CALLED FROM:   Default_Routine, this file
* ARGUMENTS:     
*     Argument             Type    IO   Description
*     --------             ----    --   -----------
*     intermediate_value    int    I    
* RETURNS:       unsigned char
*******************************************************************************/
int Limit_Mix (int intermediate_value)
{
  static int limited_value;
  
  if (intermediate_value < -127)
  {
    limited_value = -127;
  }
  else if (intermediate_value > 127)
  {
    limited_value = 127;
  }
  else
  {
    limited_value = intermediate_value;
  }
  return (int) (limited_value);
}



/*******************************************************************************
* FUNCTION NAME: Continuous_Deadband
* PURPOSE:       This filter establishes a bigger deadspot for the joysticks.
*                i.e., instead of neutral being exactly 0, it is -10 to 10.
*                so the motor is sent values of -127 to -10 or 10 to 127.
*                It also accounts for the Victor deadspot.
* CALLED FROM:   User_Drive() 
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void Continuous_Deadband(int *p1)
{
#define DeadZone  5    // Joystick deadzone is at zero +/- this value
#define MAXPOWER 124   // Apply max power past this point

  if ((*p1 > -DeadZone) && (*p1 < DeadZone))
  {
    *p1 = 0;  // Neutral
  }
  else if (*p1 >= DeadZone)
  {
    *p1 -= DeadZone;
  }
  else
  {
    *p1 += DeadZone;
  }

  if (*p1 > MAXPOWER)
  {
    *p1 = 127;  // Force to max power
  }
  else if (*p1 < -MAXPOWER)
  { 
    *p1 = -127;  // Force to max power
  }
}


/*******************************************************************************
* FUNCTION NAME: Smooth_Joystick
* PURPOSE:       This filter establishes a bigger deadspot for the joysticks.
*                i.e., instead of neutral being exactly 0, it is -10 to 10.
*                so the motor is sent values of -127 to -10 or 10 to 127.
*                It also avoids the Victor deadspot.
* CALLED FROM:   User_Drive() 
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
const rom signed char Joystick_Curve[128] = 
{
0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2,
2, 2, 3, 3, 3, 3, 4, 4, 5, 5, 5, 6, 6, 7, 7, 8,
8, 9, 9, 10, 10, 11, 11, 12, 13, 13, 14, 14, 15, 16, 17, 17,
18, 19, 20, 20, 21, 22, 23, 24, 25, 25, 26, 27, 28, 29, 30, 31,
32, 33, 34, 35, 36, 37, 38, 39, 41, 42, 43, 44, 45, 46, 48, 49,
50, 51, 53, 54, 55, 56, 58, 59, 61, 62, 63, 65, 66, 68, 69, 71,
72, 74, 75, 77, 78, 80, 81, 83, 85, 86, 88, 89, 91, 93, 95, 96,
98, 100, 102, 103, 105, 107, 109, 111, 113, 114, 116, 118, 120, 124, 127, 127, 127
};

void Smooth_Joystick(int *p1)
{
  int temp;
  temp = abs(*p1);
  temp = Joystick_Curve[temp];

  if(*p1 < 0)
    temp = -temp;

  *p1 = temp;
}


/*******************************************************************************
* FUNCTION NAME: Hold_Position
* PURPOSE:       Apply motor power to hold position if necessary.
* CALLED FROM:   User_Drive() 
*******************************************************************************/
#define HOLD_PWM 10

void Hold_Position(char init)
{
  static signed long HoldLeft, HoldRight;

  if (init)
  { // Hold robot at these encoder values
    HoldRight = Right_Encoder_Count;
    HoldLeft = Left_Encoder_Count;
  }
  else
  {
    if (HoldRight > Right_Encoder_Count)
    {
      WHL_R1 = HOLD_PWM;
    }
    else if (HoldRight < Right_Encoder_Count)
    {
      WHL_R1 = -HOLD_PWM;
    }

    if (HoldLeft > Left_Encoder_Count)
    {
      WHL_L1 = HOLD_PWM;
    }
    else if (HoldLeft < Left_Encoder_Count)
    {
      WHL_L1 = -HOLD_PWM;
    }
  }
}
