 /*******************************************************************************
* FILE NAME: sensor_routines.c
*
* DESCRIPTION:
*  This file contains sensor related routines.
*
* USAGE:
*
*******************************************************************************/

#include "ifi_utilities.h"
#include "ifi_aliases.h"
#include "ifi_default.h"

#include "user_routines.h"
#include "auto_routines.h"
#include "motor_routines.h"
#include "clock_routines.h"
#include "print_routines.h"

#define VARIABLE_DECLARE  /* causes certain global variables to be declared only once */
  #include "sensor_routines.h"
#undef VARIABLE_DECLARE

#define END_STARTUP 30   // (25)Sample gyro over n cycles to obtain accurate GyroNeutral
#define SETTLE_STARTUP 5  // Allow the gyro to settle down for n cycles after power-up
#define SAMPLE_RATE 40/250   // Sample rate of the gyro - gyro provides samples at 40Hz, we check at 250Hz

/*******************************************************************************
* FUNCTION NAME: Gyro
* PURPOSE:       Initializes and accumulates the gyro
* CALLED FROM:   user_routines_fast.c
* NOTES:
*        The gyro has a range of 1007 to 3 and a neutral point of 519-521
*        
*******************************************************************************/
void Gyro()
{
  static long LastGyro;
  long CurrentGyro;
  static unsigned long LastClocksec=0, LastClockms=0;
  static unsigned int GyroStartup=0;
  static long GyroNeutral=0;


  if (GyroTicks > 0)
  {
    /** Update gyro **/
    INTCONbits.GIEL = 0;	/* Disable Low Priority Interrupts */
    GyroTicks--;		/* Avoid the problem of decrementing it during an interrupt */
    INTCONbits.GIEL = 1;	/* Enable Low Priority Interrupts */

//    if (GyroTicks > 2)
//    {
//      PrintString((char *) " ***WARNING*** Gyro falling behind ");
//      PrintDecimal(GyroTicks);
//      PrintCR();
//    }

    if (GyroStartup == END_STARTUP)
    {
      CurrentGyro = (long)Get_Analog_Value(rc_ana_in01) - GyroNeutral;
      if (min_gyro > CurrentGyro) min_gyro = CurrentGyro;
      if (max_gyro < CurrentGyro) max_gyro = CurrentGyro;

      CurrentGyro = CurrentGyro * SAMPLE_RATE; // Add decimal points of accuracy
      GyroSum += (CurrentGyro + LastGyro) / 2;  // Is this helping or hurting?
//      GyroSum += CurrentGyro;
      LastGyro = CurrentGyro;

//      if (LastClocksec < Clocksec)
//      {
//        LastClocksec = Clocksec;
//        PrintString((char *) ", CurrentGyro = ");
//        PrintSignedDecimal(CurrentGyro);
//        PrintString((char *) ", GyroSum = ");
//        PrintSignedDecimal(GyroSum);
//        PrintString((char *) ", MinGyro = ");
//        PrintSignedDecimal(min_gyro);
//        PrintString((char *) ", MaxGyro = ");
//        PrintSignedDecimal(max_gyro);
//        PrintCR();
//      }
    }
    else /** Initializing Gyro **/
    {
      GyroStartup++;
      /** Sample over n cycles to obtain accurate GyroNeutral **/
      /** Allow n cycles for the gyro to be properly powered-up and settle down **/
      if (GyroStartup > SETTLE_STARTUP)
      {
        GyroNeutral += Get_Analog_Value(rc_ana_in01)<<6; // Increase accuracy
        if (GyroStartup == END_STARTUP)
        {
          GyroNeutral /= (END_STARTUP-SETTLE_STARTUP);
          GyroNeutral = (GyroNeutral+32)>>6; // Round to original accuracy
          min_gyro = max_gyro = 0;
          LastGyro = 0;
          GyroSum = 0;

//          PrintString((char *) " ***GyroNeutral = ");
//          PrintDecimal(GyroNeutral);
//          PrintCR();
        }
      }
    }
  }
}

/*******************************************************************************
* FUNCTION NAME: Gyro_Degrees
* PURPOSE:       Convert gyrosum to degrees in the range -180 to 180
* NOTES:         By definition 0 degrees is the original robot heading
*          
*******************************************************************************/
  #define DEG90   8700  // Gyro ticks in  90 degrees
  #define DEG180 17400  // Gyro ticks in 180 degrees
  #define DEG360 34800  // Gyro ticks in 360 degrees

int Gyro_Degrees()
{
  long temp;

  temp = GyroSum % DEG360;

  if (temp > DEG180)
    temp = (DEG360 - temp) * -1;
  else if (temp < -DEG180)
    temp = DEG360 + temp;
  else if (temp == -DEG180)
    temp = DEG180;

  return temp * 180 / DEG180;
}


/*******************************************************************************
* FUNCTION NAME: Current_Sensor
* PURPOSE:       Monitors the current sensors
* CALLED FROM:   user_routines_fast.c
*******************************************************************************/
#define VOLTS_MAX 10        // Number of readings to average over
#define MID_ANALOG_RANGE 512  // analog value representing 0 volts

void Current_Sensor()
{
  static int LeftVoltsArray[VOLTS_MAX] = {0,0,0,0,0,0,0,0,0,0};
  static char LeftVoltsIndex=0;
  static int RightVoltsArray[VOLTS_MAX] = {0,0,0,0,0,0,0,0,0,0};
  static char RightVoltsIndex=0;
  int LeftVolts, RightVolts;  // volts are in terms of analog units only


  LeftVolts  = (int)Get_Analog_Value(rc_ana_in02) - MID_ANALOG_RANGE; // 0= 0 volts
  RightVolts = (int)Get_Analog_Value(rc_ana_in03) - MID_ANALOG_RANGE;
  
  LeftVolts = Average_Values(&LeftVoltsIndex, LeftVolts, (char)VOLTS_MAX, LeftVoltsArray);
  RightVolts = Average_Values(&RightVoltsIndex, RightVolts, (char)VOLTS_MAX, RightVoltsArray);

  /* Convert analog value to amps */
  LeftAmps  = LeftVolts  * 75 / 306;  // Convert analog value to amps ([818.4 - 512] analog = 75 amps)
  RightAmps = RightVolts * 75 / 306;  // Convert analog value to amps 
  if (LeftAmps  < 0) LeftAmps  = -LeftAmps; // Account for running in reverse, we don't need the sign
  if (RightAmps < 0) RightAmps = -RightAmps;
}



/*******************************************************************************
* FUNCTION NAME: Average_Values
* PURPOSE:       Return an average of the last "max" values
* CALLED FROM:   user_routines_fast.c
*******************************************************************************/
int Average_Values(char *index, int newvalue, char max, int *array)
{
  int i, avg = 0;

  array[*index] = newvalue;
  *index = *index + 1;
  if ((*index) == max)
    *index = 0;

  for (i=0; i<max;i++)
    avg += array[i];
  avg /= max;

  return avg;
}


/*******************************************************************************
* FUNCTION NAME: Encoder_PID
* PURPOSE:       
* CALLED FROM:   user_routines_fast.c
*   gyro correction, IR correction, encoder correction, regular driving
*   !!!!!!!!!!!!!!Not debugged!!!!!!!!!!!
*   Max free wheeling speed of encoders is ~2575 counts/sec or ~9.43 fps
*******************************************************************************/
#define PP 1/10  // Convert error to a PWM - basic proportional power adjustment
#define II 1  // Convert error to a PWM - cumulative error increases power adjustment
#define DD 1  // Convert error to a PWM - sudden reaction to sudden changes

#define FULLENCODERS 1024  // 4 full revs = 3.75 feet
#define POWEROFFSET 10     // Always apply at least a minimum amount of power

int Encoder_PID(int expected, int actual, int *LastActual, int *CumError)
{
  int pwm;

  /* Proportional */
  pwm = FULLSPEED * (expected - actual) + POWEROFFSET / FULLENCODERS; // pwr as % of full power

  /* Derivative */
//  pwm -=  (actual - *LastActual) / DD;

  /* Integral */
  *CumError += expected - actual;
  pwm += *CumError * FULLSPEED / FULLENCODERS;

  *LastActual = actual;

  if (pwm > FULLSPEED) pwm = FULLSPEED;
  if (pwm < -FULLSPEED) pwm = -FULLSPEED;

  return pwm;
}


int Gyro_Degrees(void);

/*******************************************************************************
* FUNCTION NAME: Gyro_PID
* PURPOSE:       
* CALLED FROM:   user_routines_fast.c
*   gyro correction, IR correction, encoder correction, regular driving
*******************************************************************************/
#define CUMFACTOR 0   // Amplify the cumulative error
#define PWRFACTOR 35  // Amplify the current error

void Gyro_PID(int expected,char init)
{
  int actual;
  static int lastActual, cumError;
  int correction;

  actual = Gyro_Degrees();
  /** Convert gyrodegrees to be as close as possible to heading **/
  /*  This avoids issues with wrapping over the 180 degree mark (-179 + 1 = 180)*/
  if (actual > (expected+180)) actual -= 360;
  if (actual < (expected-180)) actual += 360;

  if (init==TRUE)
  {
    cumError = 0;
    lastActual = actual;
    PrintString((char *) " Gyro_PID init ");
  }
  else
  {
     if (expected > actual)
      cumError += expected - actual;
     else
      cumError += actual - expected;

    /* WHL_L1 & WHL_R1 are assumed to be identical for straight travel */
    correction = WHL_L1 * (actual - expected)/PWRFACTOR;

    PrintString((char *) ", pwr correction=");
    PrintSignedDecimal(correction);
    correction += cumError * CUMFACTOR;
    PrintString((char *) ", cum correction=");
    PrintSignedDecimal(correction);
    PrintCR();

    if (WHL_L1 > 0) // Driving backward
    {
     if (expected > actual)
        WHL_R1 -= correction;
      else
        WHL_L1 -= correction;
    }
    else  // Driving backward
    {
      if (expected > actual)
        WHL_L1 += correction;
      else
        WHL_R1 += correction;
    }
  }
}


/*******************************************************************************
*	FUNCTION:		Initialize_Interrupts()
*	PURPOSE:		Initializes the interrupt hardware.
*	CALLED FROM:	user_routines.c/User_Initialization()
*******************************************************************************/
void Initialize_Interrupts(void)  
{
	// initialize external interrupt 1 (INT2 on user 18F8520)
	TRISBbits.TRISB2 = 1;		// make sure the RB2/INT2 pin is configured as an input [108]
								//
	INTCON3bits.INT2IP = 0;		// 0: interrupt 1 is low priority (leave at 0 for IFI controllers) [91]
								// 1: interrupt 1 is high priority
								//
	INTCON2bits.INTEDG2 = 1;	// 0: trigger on the falling-edge [90]
								// 1: trigger on the rising-edge
								//
	INTCON3bits.INT2IE = 1;		// 0: disable interrupt	1 [91]
								// 1: enable interrupt 1
	
	// initialize external interrupt 2 (INT3 on user 18F8520)
	TRISBbits.TRISB3 = 1;		// make sure the RB3/CCP2/INT3 pin is configured as an input [108]
								//
	INTCON2bits.INT3IP = 0;		// 0: interrupt 2 is low priority (leave at 0 for IFI controllers) [90]
								// 1: interrupt 2 is high priority
								//
	INTCON2bits.INTEDG3 = 1;	// 0: trigger on the falling-edge [90]
								// 1: trigger on the rising-edge
								//
	INTCON3bits.INT3IE = 1;		// 0: disable interrupt	2 [91]
								// 1: enable interrupt 2
}
